/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Controller.hpp"
#include "MapDisplay.hpp"
#include "GameWindow.hpp"

#include "Fractal.hpp"
#include "Figure.hpp"

#include <gdk/gdkkeysyms.h>

#include <cmath>

namespace islandmandel
{
namespace gimandel
{

/* ************************************************************************** */
/* **************************** Controller ********************************** */
/* ************************************************************************** */

/** Amount to move fractal for arrow keys */
#define MOVE_ARROW 100
/** Scaling factor for plus/minus keys */
#define SCALING_PLUSMINUS 4.

Controller::Controller(GameWindow& g, MapDisplay& m)
  : game(g), map(m)
#ifndef NDEBUG
  , fract(NULL), fig(NULL)
#endif
{
  // Enable events
  gtk_widget_add_events(map.getWidget(), GDK_KEY_RELEASE_MASK);
  gtk_widget_add_events(map.getWidget(), GDK_BUTTON_MOTION_MASK);
  gtk_widget_add_events(map.getWidget(), GDK_BUTTON_PRESS_MASK);
  gtk_widget_add_events(map.getWidget(), GDK_BUTTON_RELEASE_MASK);
  GTK_WIDGET_FLAGS(map.getWidget())|=GTK_CAN_FOCUS;
  gtk_widget_grab_focus(map.getWidget());

  // Install the handlers
  g_signal_connect(G_OBJECT(map.getWidget()), "key_release_event",
    G_CALLBACK(&handle_keyPress), this);
  g_signal_connect(G_OBJECT(map.getWidget()), "button_press_event",
    G_CALLBACK(&handle_mousePress), this);
  g_signal_connect(G_OBJECT(map.getWidget()), "button_release_event",
    G_CALLBACK(&handle_mouseRelease), this);
  g_signal_connect(G_OBJECT(map.getWidget()), "motion_notify_event",
    G_CALLBACK(&handle_mouseDrag), this);
}

// Code in the handlers to revive the Controller from the gpointer
#define REVIVE_CONTROLLER \
  Controller& me(*static_cast<Controller*>(ptr)); \
  assert(w==me.map.getWidget());

// Code used in the mouse handlers to construct the position in our Coordinates
#define DEFINE_CLICK_POSITION \
  const Coordinate clickPos( \
    Coordinate::round(evt->x), \
    Coordinate::round(evt->y));

// Code to check the modifier keys
#define CHECK_MODIFIER_CTRL \
  const bool ctrl(evt->state & GDK_CONTROL_MASK);
#define CHECK_MODIFIER_SHIFT \
  const bool shift(evt->state & GDK_SHIFT_MASK);
  
// Button-code for left/right buttons
#define BUTTON_LEFT 1
#define BUTTON_RIGHT 3

gboolean Controller::handle_mousePress(GtkWidget* w, GdkEventButton* evt,
  gpointer ptr)
{
  REVIVE_CONTROLLER
  DEFINE_CLICK_POSITION
  CHECK_MODIFIER_CTRL

  /* Mouse button presses do:
   * - right button: store position so scaling/rotation can be tracked
   */

  // Right button, no ctl:  Start rotation/scaling
  if(!ctrl && evt->button==BUTTON_RIGHT)
  {
    me.rotscalStart=clickPos;
    return TRUE;
  }

  // Nothing interesting otherwise
  return FALSE;
}

gboolean Controller::handle_mouseRelease(GtkWidget* w, GdkEventButton* evt,
  gpointer ptr)
{
  REVIVE_CONTROLLER
  DEFINE_CLICK_POSITION
  CHECK_MODIFIER_CTRL
  CHECK_MODIFIER_SHIFT

  /* Mouse button releases can do:
   * - Ctrl-shift zooms in/out with left/right button, centers first
   * - Ctrl centers fractal on position
   * - ordinary click moves figure there
   * - ordinary right click finishes rotation/scaling
   */

  // Zoom factor if applicable
  float fact(1.);

  // Ctrl-shift-clicks zoom the fractal
  if(ctrl && shift)
  {
    if(evt->button==1)
      fact*=SCALING_PLUSMINUS;
    else
      fact/=SCALING_PLUSMINUS;
  }

  // Ctrl-[Shift-]click centers fractal, possibly zooming as well
  if(ctrl)
  {
    const Coordinate::compT mx(clickPos.getX()-me.fract->getWidth()/2);
    const Coordinate::compT my(clickPos.getY()-me.fract->getHeight()/2);

    if(fact==1.)
      me.fract->move(mx, my);
    else
      me.fract->moveZoom(mx, my, fact);

    me.map.redraw();
    return TRUE;
  }

  // Ordinary left click moves Figure
  assert(!ctrl);
  if(evt->button==BUTTON_LEFT)
  {
    assert(me.fig);
    me.fig->setOffset(clickPos);
    me.map.redraw();
    return TRUE;
  }

  // Ordinary right click finishes rotation/scaling
  if(evt->button==BUTTON_RIGHT)
  {
    me.doRotScal(clickPos);
    assert(me.fig);
    me.fig->acceptTransform();
    me.map.redraw();
    return TRUE;
  }

  // Otherwise uninteresting
  return FALSE;
}

gboolean Controller::handle_mouseDrag(GtkWidget* w, GdkEventMotion* evt,
  gpointer ptr)
{
  REVIVE_CONTROLLER
  DEFINE_CLICK_POSITION
  CHECK_MODIFIER_CTRL

  /* Mouse drags (move while button down) can do:
   * - if left button, move the figure along in "real time"
   * - if right button, rotate/scale on the fly
   */

  // Ctrl-clicks are uninteresting
  if(ctrl)
    return FALSE;

  // Left button, move the fractal
  if(evt->state & GDK_BUTTON1_MASK)
  {
    assert(me.fig);
    me.fig->setOffset(clickPos);
    me.map.redraw();
    return TRUE;
  }

  // Right button, scale/rotate
  if(evt->state & GDK_BUTTON3_MASK)
  {
    me.doRotScal(clickPos);
    me.map.redraw();
    return TRUE;
  }
  
  // Otherwise, ignore
  return FALSE;
}

gboolean Controller::handle_keyPress(GtkWidget* w, GdkEventKey* evt,
  gpointer ptr)
{
  REVIVE_CONTROLLER

  /* Keyboard UI:
   * - arrows move fractal by fixed amount
   * - +/- keys zoom in/out
   * - p/n/r do previous/next/restart level
   * - enter does next level if figure is placed correctly
   * - c shows the configuration dialog again
   */

  switch(evt->keyval)
  {

    // Arrows move the fractal
    case GDK_Left:
      me.fract->move(-MOVE_ARROW, 0);
      break;
    case GDK_Right:
      me.fract->move(MOVE_ARROW, 0);
      break;
    case GDK_Up:
      me.fract->move(0, -MOVE_ARROW);
      break;
    case GDK_Down:
      me.fract->move(0, MOVE_ARROW);
      break;

    // Scale using +/-
    case GDK_plus:
      me.fract->zoom(SCALING_PLUSMINUS);
      break;
    case GDK_minus:
      me.fract->zoom(1./SCALING_PLUSMINUS);
      break;

    // Handle the level controlling commands
    case GDK_p:
      me.game.prevLevel();
      break;
    case GDK_n:
      me.game.nextLevel();
      break;
    case GDK_r:
      me.game.restart();
      break;

    // Check figure placement and proceed to next level
    case GDK_Return:
    {
      bool ok(true);
      assert(me.fig);
      for(Figure::Iterator i(*me.fig); i; ++i)
        if(i.getColour()!=me.fract->getPixelColour(i.getCoord()))
        {
          ok=false;
          break;
        }
      if(ok)
        me.game.nextLevel();
      break;
    }

    // Reconfigure
    case GDK_c:
      me.game.reconfig();
      break;

    // Unknown key
    default:
      return FALSE;

  }

  me.map.redraw();
  return TRUE;
}

void Controller::setFigure(Figure& f)
{
  fig=&f;
  assert(fig);

  fig->setOffset(Coordinate(fract->getWidth()/2, fract->getHeight()/2));
}

void Controller::doRotScal(const Coordinate& pos)
{
  assert(fig);

  // We need difference to figure-position
  const Coordinate relStart(rotscalStart-fig->getOffset());
  const Coordinate relStop(pos-fig->getOffset());

  // Now, scale accordingly to relation of both norms
  const Transformation::scaleFactT f(relStop.norm()/relStart.norm());

  // And find the angle-difference
  const float angStart(std::atan2(relStart.getY(), relStart.getX()));
  const float angStop(std::atan2(relStop.getY(), relStop.getX()));
  const Transformation::angleT ang(angStop-angStart);

  // Perform it
  const Transformation scal(Transformation::forScaling(f));
  const Transformation rot(Transformation::forRotation(ang));
  fig->setTransform(rot*scal);
}

} // namespace gimandel
} // namespace islandmandel
