/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "FractalDisplay.hpp"

#include "CSet.hpp"
#include "Checker.hpp"

namespace islandmandel
{
namespace gimandel
{

/* ************************************************************************** */
/* ************************** FractalDisplay ******************************** */
/* ************************************************************************** */

FractalDisplay::FractalDisplay()
  : canv(gtk_drawing_area_new()), fract(NULL), colourMap(NULL)
{
  assert(canv);

  // Register expose event
  g_signal_connect(G_OBJECT(canv), "expose-event",
    G_CALLBACK(&on_expose), this);

  // Rest is done in setFractal(const Fractal&)
}

FractalDisplay::~FractalDisplay()
{
  if(colourMap)
    gdk_rgb_cmap_free(colourMap);
}

void FractalDisplay::setFractal(const Fractal& fr)
{
  if(colourMap)
    gdk_rgb_cmap_free(colourMap);

  // Take over the new reference
  fract=&fr;
  assert(fract);

  // Adapt size
  gtk_widget_set_size_request(canv, fract->getWidth(), fract->getHeight());

  // Create the colour-map depending on fractal-type
  guint32 colours[256];
  if(dynamic_cast<const CSet*>(fract))
  {
    assert(fract->getNumberColours()==2);
    colours[0]=0x000000;
    colours[1]=0xFFFFFF;
  } else if(dynamic_cast<const Checker*>(fract))
  {
    assert(fract->getNumberColours()>=2 && fract->getNumberColours()<=3);
    colours[0]=0x000000;
    colours[1]=0xFFFFFF;
    colours[2]=0x777777;
  }
#ifndef NDEBUG
  else
    assert(false);
#endif
  colourMap=gdk_rgb_cmap_new(colours, fract->getNumberColours());
  assert(colourMap);
}

void FractalDisplay::redraw() const
{
  gdk_window_invalidate_rect(canv->window, NULL, FALSE);
}

void FractalDisplay::draw(GdkDrawable* screen, GdkGC* gc) const
{
  // Draw the fractal
  assert(fract); // Should have been set prior to painting!
  assert(colourMap);
  gdk_draw_indexed_image(screen, gc,
    0, 0, fract->getWidth(), fract->getHeight(),
    GDK_RGB_DITHER_NORMAL,
    const_cast<guchar*>(fract->getPixels()), fract->getRowstride(),
    colourMap);
}

void FractalDisplay::draw() const
{
  // Our drawable is the canvas itself
  GdkDrawable* screen(canv->window);
  assert(screen);

  // Create the GC
  GdkGC* gc(gdk_gc_new(screen));
  assert(gc);

  // Do the actual drawing
  draw(screen, gc);

  // Free the GC
  g_object_unref(G_OBJECT(gc));
}

gboolean FractalDisplay::on_expose(GtkWidget* w, GdkEventExpose* evt, 
  gpointer ptr)
{
  const FractalDisplay& me(*static_cast<const FractalDisplay*>(ptr));
  assert(w==me.canv);
  me.draw();
  return TRUE;
}

} // namespace gimandel
} // namespace islandmandel
