/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef GIMANDEL_GAMEWINDOW_HPP
#define GIMANDEL_GAMEWINDOW_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "MapDisplay.hpp"
#include "Controller.hpp"
#include "ConfigDialog.hpp"

#include "Figure.hpp"

#include <gtk/gtk.h>

#include <cassert>

namespace islandmandel
{
class Fractal;
namespace gimandel
{

/* ************************************************************************** */
/* **************************** GameWindow ********************************** */
/* ************************************************************************** */

/**
 * The main window, here takes the actual game place; owns the Controller and
 * MapDisplay.
 */
class GameWindow
{

  /* ************************************************************************ */
  public:

    /**
     * Construct the window, has to be configured before first use.
     */
    GameWindow();

    /**
     * Destroy, need to free the fractal/figure
     */
    ~GameWindow();

    /**
     * Show the window, poping up a configuration window first.
     * @return False iff configuration was cancelled, quit.
     */
    bool show();

    /**
     * Request a re-configuration; show the dialog and take appropriate action
     * should something change.
     */
    inline void reconfig()
    {
      if(configDlg.run())
        map.redraw();
    }

    // Handle figure-changes
    inline void restart()
    {
      constructFigure(true);
    }
    inline void nextLevel()
    {
      ++figDots;
      constructFigure(true);
    }
    inline void prevLevel()
    {
      if(figDots>1)
        --figDots;
      assert(figDots>=1);
      constructFigure(true);
    }

    /**
     * Set/exchange the fractal in use
     * @param f The new fractal, ownership is taken
     */
    void setFractal(Fractal* f);

    /**
     * Access the Gtk+ widget
     * @return The Gtk+ widget backing
     */
    inline GtkWidget* getWidget()
    {
      assert(w);
      return w;
    }

  /* ************************************************************************ */
  private:

    /** Our fractal, owned by us */
    Fractal* fract;

    /** Number of points for the figure currently */
    Figure::pointNumT figDots;

    /** The figure used */
    Figure* fig;

    /** The MapDisplay */
    MapDisplay map;
    /** The Controller */
    Controller control;

    /** The Gtk+ window widget */
    GtkWidget* w;

    /** The configuration dialog used */
    ConfigDialog configDlg;

    /**
     * Construct a new figure with the correct configuration
     * @param refresh Whether to refresh the drawing or not
     */
    void constructFigure(bool refresh);

    /**
     * Signal handler for closing the window, this shuts down the application.
     * @param w The widget
     * @param ptr this pointer
     */
    static void handle_destroy(GtkWidget* w, gpointer ptr);

};

} // namespace gimandel
} // namespace islandmandel

#endif // Header-guard
