/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "Figure.hpp"

#include <cmath>
#include <cstdlib>

namespace islandmandel
{

/* ************************************************************************** */
/* ***************************** Figure ************************************* */
/* ************************************************************************** */

Figure::Figure(pointNumT len, Coordinate::compT r, Coordinate::compT d,
  colourT cols)
  : n(len), pts(new Dot[n])
{
  assert(pts);
  assert(n>=1);
  assert(r>0);
  assert(d>=0);

  // Randomly choose points
  pointNumT chosen(0);
  while(chosen!=n)
  {
    assert(chosen<n);

    // Random angle
    const float alpha(2*M_PI*std::rand()/RAND_MAX);

    /* r is chosen roughly normal-distributed by summing three random numbers.
     * rand()+rand()+rand() -> bell-shaped form in (0, 3*RAND_MAX)
     * /1.5*RAND_MAX -> bell-shaped form in (0, 2)
     * *r -> bell-shaped form in (0, 2*r), mission accomplished.
     */
    float curR(static_cast<float>(std::rand())+std::rand()+std::rand());
    curR/=1.5*RAND_MAX;
    curR*=r;
    assert(curR>0 && curR<2*r);

    // Calculate the coordinates
    pts[chosen].first=Coordinate(curR*std::cos(alpha), curR*std::sin(alpha));

    // Check if it is far enough away from the other points
    bool ok(true);
    for(const Dot* i(pts); i!=pts+chosen; ++i)
    {
      const Coordinate::compT dx(i->first.getX()-pts[chosen].first.getX());
      const Coordinate::compT dy(i->first.getY()-pts[chosen].first.getY());
      if(d*d>dx*dx+dy*dy)
      {
        ok=false;
        break;
      }
    }

    // If yes, accept it and choose colour
    if(ok)
    {
      pts[chosen].second=static_cast<colourT>(std::rand()%cols);
      ++chosen;
    }
  }
  assert(chosen==n);
}

Figure::~Figure()
{
  assert(pts);
  delete[] pts;
}

void Figure::acceptTransform()
{
  // Transform the points
  for(pointNumT i(0); i!=n; ++i)
    pts[i].first=transf(pts[i].first);

  // Reset to identity
  setTransform(Transformation());
}

} // namespace islandmandel
