/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef IMANDEL_FIGURE_HPP
#define IMANDEL_FIGURE_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include "Geometry.hpp"

#include <cassert>
#include <utility>

namespace islandmandel
{

/* ************************************************************************** */
/* ***************************** Figure ************************************* */
/* ************************************************************************** */

/**
 * A dot-figure to be placed atop the fractal.  It consists of some dots
 * placed around (0, 0) with random angles and a radius normal-distributed.
 *
 * Together with the points, the modifications of the Figure (moving, scaling,
 * rotating) are stored as an offset for the movements, and a transformation
 * for scaling/rotating; once a scale/rotate-modification is done (finally
 * accepted by the user), the points are transformed accordingly.
 */
class Figure
{

  /* ************************************************************************ */
  public:

    /** Type for number of points */
    typedef unsigned pointNumT;

    /** Type used as colour-code */
    typedef unsigned char colourT;

    /** (Constant) iterator used through the (transformed) points */
    class Iterator;

    /**
     * Construct a new figure with given characteristics
     * @param len The number of points to choose
     * @param r The mean radius of the requested circle
     * @param d Minimum distance of two points to avoid collision
     * @param cols Number of colours to choose from
     */
    Figure(pointNumT len, Coordinate::compT r, Coordinate::compT d,
      colourT cols);

    /**
     * Destroy, freeing the point-array
     */
    ~Figure();

    /**
     * Set transformation in progress, fully replacing the current one
     * @param t The new transformation
     */
    inline void setTransform(const Transformation& t)
    {
      transf=t;
    }

    /**
     * Set the offset.
     * @param o The new offset
     */
    inline void setOffset(const Coordinate& o)
    {
      offset=o;
    }

    /**
     * Retrieve the offset
     * @return Our offset
     */
    inline const Coordinate& getOffset() const
    {
      return offset;
    }

    /**
     * Accept the currently set transformation finally; this modifies the
     * points themselves accordingly and resets the transformation to the
     * identity.
     */
    void acceptTransform();

  /* ************************************************************************ */
  private:

    /** Number of points */
    const pointNumT n;

    /** A point in our map, coordinate and colour */
    typedef std::pair<Coordinate, colourT> Dot;
    /** The points themselves */
    Dot* pts;

    /** The Transformation in progress */
    Transformation transf;

    /** The offset to apply to the points */
    Coordinate offset;

};

/* ************************************************************************** */
/* ******************************* Figure::Iterator ************************* */
/* ************************************************************************** */

/**
 * Iterator through the points of a Figure, applying its transformation.
 */
class Figure::Iterator
{

  /* ************************************************************************ */
  public:

    /**
     * Construct an Iterator for a Figure
     * @param f The figure to iterate through
     */
    explicit inline Iterator(const Figure& f)
      : fig(f), i(0)
    {}

    /**
     * Check if we are at the end
     * @return True iff there are more points
     */
    inline operator bool() const
    {
      assert(i<=fig.n);
      return i!=fig.n;
    }

    // Access next point's coordinate/colour
    inline Coordinate getCoord() const
    {
      assert(operator bool());
      return fig.transf(fig.pts[i].first)+fig.offset;
    }
    inline Figure::colourT getColour() const
    {
      assert(operator bool());
      return fig.pts[i].second;
    }

    /**
     * Increment to next point
     */
    inline void operator++()
    {
      ++i;
    }

  /* ************************************************************************ */
  private:

    /** The Figure we belong to */
    const Figure& fig;

    /** Index of next point to return */
    Figure::pointNumT i;

};

} // namespace islandmandel

#endif // Header-guard
