/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef IMANDEL_GEOMETRY_HPP
#define IMANDEL_GEOMETRY_HPP

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <cmath>

/* Geometry in imandel is based on integer-coordinates (pixels), at least what
 * "comes out".  Operation-matrices for scaling/rotation are made up from
 * floats, but whenever they are applied to real coordinates those are again
 * rounded afterwards.  Coordinates are signed.
 */

namespace islandmandel
{

/* ************************************************************************** */
/* ***************************** Coordinate ********************************* */
/* ************************************************************************** */

/**
 * A 2D coordinate of coordT's, supports addition/subtraction like vectors for
 * dealing with offsets.
 */
class Coordinate
{

  /* ************************************************************************ */
  public:

    /** The type used for coordinate-components */
    typedef int compT;

    // Create, possibly default to zero for when we need an array/vector
    inline Coordinate()
      : x(0), y(0)
    {}
    inline Coordinate(compT xc, compT yc)
      : x(xc), y(yc)
    {}
    inline Coordinate(float xc, float yc)
      : x(round(xc)), y(round(yc))
    {}

    // Default copy-construction and assignment is ok
    
    // Accessors to components
    inline compT getX() const
    {
      return x;
    }
    inline compT getY() const
    {
      return y;
    }

    /**
     * Calculate the norm
     * @return This coordinate's norm
     */
    inline float norm() const
    {
      return std::sqrt(x*x+y*y);
    }

    // Comparison
    friend inline bool operator==(const Coordinate& a, const Coordinate& b)
    {
      return a.x==b.x && a.y==b.y;
    }
    friend inline bool operator!=(const Coordinate& a, const Coordinate& b)
    {
      return !(a==b);
    }

    /**
     * Sum two coordinates component-wise (vector and offset)
     * @param a The first vector
     * @param b The second vector
     * @return The sum vector
     */
    friend inline Coordinate operator+(const Coordinate& a, const Coordinate& b)
    {
      return Coordinate(a.x+b.x, a.y+b.y);
    }
    friend inline Coordinate operator-(const Coordinate& a, const Coordinate& b)
    {
      return Coordinate(a.x-b.x, a.y-b.y);
    }

    /**
     * Function to do the coordinate rounding
     * @param c The coordinate-value as entryT to round
     * @return The rounded value as Coordinate::compT
     */
    static inline Coordinate::compT round(float c)
    {
      return static_cast<Coordinate::compT>(c+(c<0 ? -.5 : .5));
    }


  /* ************************************************************************ */
  private:

    /** The x-component */
    compT x;

    /** The y-component */
    compT y;

};

/* ************************************************************************** */
/* ******************************* Transformation *************************** */
/* ************************************************************************** */

/**
 * Describes a coordinate-transformation apart from offset-addition, that is,
 * rotation and scaling.  It is simply a matrix internally.
 */
class Transformation
{

  /* ************************************************************************ */
  public:

    /** Type used as scaling factor */
    typedef float scaleFactT;
    /** Type used as angle */
    typedef float angleT;

    /**
     * The default constructor creates the identity transformation
     */
    inline Transformation()
    {
      set(1, 0, 0, 1);
    }

    // Default copy-constructor and assignment is ok if needed

    /**
     * Apply the transformation to a coordinate, this means
     * matrix-vector-product.
     * @param c The coordinate to transform
     * @return The transformed coordinate
     */
    Coordinate operator()(const Coordinate& c) const;

    /**
     * Chain two transformations to a new one.
     * @param a The one applied afterwards
     * @param b The one applied prior
     * @return The chained transformation
     */
    friend Transformation operator*(const Transformation& a,
      const Transformation& b);

    // Create for certain operations
    static Transformation forScaling(scaleFactT f);
    static Transformation forRotation(angleT alpha);

  /* ************************************************************************ */
  private:

    /** Type used for the matrix-entries */
    typedef float entryT;

    /** The matrix as 2x2 */
    entryT matrix[2][2];

    /**
     * Create the matrix for the entries [a b; c d]
     * @param a The entry a11
     * @param b The entry a12
     * @param c The entry a21
     * @param d The entry a22
     */
    inline Transformation(entryT a, entryT b, entryT c, entryT d)
    {
      set(a, b, c, d);
    }

    /**
     * Directly set all four components of the matrix
     * @param a The entry a11
     * @param b The entry a12
     * @param c The entry a21
     * @param d The entry a22
     */
    inline void set(entryT a, entryT b, entryT c, entryT d)
    {
      matrix[0][0]=a;
      matrix[0][1]=b;
      matrix[1][0]=c;
      matrix[1][1]=d;
    }

};

} // namespace islandmandel

#endif // Header-guard
