/*

    Island Mandel, a puzzle game taking place on the Mandelbrot-set
    Copyright (C) 2008,2012  Daniel Kraft <d@domob.eu>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

/* Test the creation of random figures statistically; create some, check for
 * absolute criteria like minimum distance between each two points, and do a
 * rough check that the majority of radii is in (r/2, 3*r/2), but some out of
 * this interval, too.  If done often enough, we can be sure this is met each
 * time the test is run.
 */

#include "Test.hpp"

#include "Geometry.hpp"
#include "Figure.hpp"

#include <cstdlib>
#include <cmath>
#include <iostream>

using namespace islandmandel;
using namespace tests;

// Figure configuration values
#define USED_R 100
#define USED_MIND 5
#define USED_LEN 10
#define USED_COLS 5
#define TRIES 1000

/**
 * Test if a figure meets the absolute criteria
 * @param f The figure to check
 */
void checkFixed(const Figure& f)
{
  unsigned num(0);
  for(Figure::Iterator i(f); i; ++i)
  {
    unsigned numSame(0);
    for(Figure::Iterator j(f); j; ++j)
    {
      if(i.getCoord()==j.getCoord())
        ++numSame;
      else
      {
        const Coordinate::compT dx(i.getCoord().getX()-j.getCoord().getX());
        const Coordinate::compT dy(i.getCoord().getY()-j.getCoord().getY());
        check(USED_MIND*USED_MIND<=dx*dx+dy*dy, "Points are too close!");
      }
    }
    check(numSame==1, "Wrong number of 'same' points in figure!");
    check(i.getColour()<USED_COLS, "Too high colour code given!");
    ++num;
  }

  check(num==USED_LEN, "Created figure has wrong number of points!");
}

int main()
{
  // We do want randomness
  std::srand(time(NULL));

  // Count number of points in and out (r/2, 3*r/2)
  unsigned numGood(0), numBad(0);

  // Create some figures and analyze them
  for(unsigned i(0); i!=TRIES; ++i)
  {
    Figure fig(USED_LEN, USED_R, USED_MIND, USED_COLS);
    checkFixed(fig);

    for(Figure::Iterator p(fig); p; ++p)
    {
      const Coordinate::compT x(p.getCoord().getX());
      const Coordinate::compT y(p.getCoord().getY());
      const Coordinate::compT r(Coordinate::round(std::sqrt(x*x+y*y)));

      if(r<USED_R/2 || r>3*USED_R/2)
        ++numBad;
      else
        ++numGood;
    }
  }

  // Good+bad must equal to total number
  check(numBad+numGood==USED_LEN*TRIES, 
    "numBad and numGood gives wrong total!");

  // Print out number of good and bad points
  std::cout << "Good points: " << numGood << std::endl;
  std::cout << "Bad points:  " << numBad << std::endl;

  // Chosen criterium for accepting: >80% good, >10% bad
  check(numGood>USED_LEN*TRIES*8/10, "Too few good points!");
  check(numBad>USED_LEN*TRIES*1/10, "Too few bad points!");

  return EXIT_SUCCESS;
}
